# -*- coding: utf-8 -*-

import re
import os
import toml
import subprocess
import warnings

try:
    import pypandoc
except ImportError:
    pypandoc = None

from .exceptions.poet import MissingElement, InvalidElement
from .version_parser import VersionParser
from .build import Builder
from .package import Dependency, PipDependency


class Poet(object):

    EXCLUDES = ()
    INCLUDES = ()

    def __init__(self, path, builder=Builder()):
        self._path = path
        self._dir = os.path.dirname(path)
        self._builder = builder
        self._git_config = None

        self._name = None
        self._version = None
        self._description = None
        self._authors = []
        self._homepage = None
        self._repository = None
        self._keywords = []
        self._python_versions = []
        self._dependencies = []
        self._dev_dependencies = []
        self._pip_dependencies = []
        self._pip_dev_dependencies = []
        self._scripts = {}
        self._license = None
        self._readme = None
        self._include = []
        self._exclude = []
        self._extensions = {}

        with open(self._path) as f:
            self._config = toml.loads(f.read())

        self.load()

    @property
    def base_dir(self):
        return self._dir

    @property
    def git_config(self):
        if self._git_config is not None:
            return self._git_config

        config_list = subprocess.check_output(
            ['git', 'config', '-l']
        ).decode()

        self._git_config = {}

        m = re.findall('(?ms)^([^=]+)=(.*?)$', config_list)
        if m:
            for group in m:
                self._git_config[group[0]] = group[1]

        return self._git_config

    @property
    def ignore(self):
        ignore_files = [
            '.gitignore'
        ]

        ignore = []

        for filename in ignore_files:
            filepath = os.path.join(self._dir, filename)
            if not os.path.exists(filepath):
                continue

            with open(filepath) as fd:
                for line in fd.readlines():
                    if re.match('^\s*#.*$', line):
                        continue

                    ignore.append(line.strip())

        return ignore

    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version

    @property
    def description(self):
        return self._description

    @property
    def authors(self):
        return self._authors

    @property
    def homepage(self):
        return self._homepage

    @property
    def repository(self):
        return self._repository

    @property
    def keywords(self):
        return self._keywords

    @property
    def python_versions(self):
        return self._python_versions

    @property
    def dependencies(self):
        return self._dependencies

    @property
    def dev_dependencies(self):
        return self._dev_dependencies

    @property
    def pip_dependencies(self):
        return self._pip_dependencies

    @property
    def pip_dev_dependencies(self):
        return self._pip_dev_dependencies

    @property
    def scripts(self):
        return self._scripts

    @property
    def license(self):
        return self._license

    @property
    def readme(self):
        return self._readme

    @property
    def include(self):
        return self._include

    @property
    def exclude(self):
        return self._exclude

    @property
    def extensions(self):
        return self._extensions

    @property
    def lock_file(self):
        return os.path.join(self._dir, 'poetry.lock')

    @property
    def lock(self):
        from .lock import Lock

        return Lock(self.lock_file)

    @property
    def path(self):
        return self._path

    @property
    def archive(self):
        return '{}-{}.tar.gz'.format(self.name, self.version)

    def load(self):
        """
        Load data from the config.
        """
        self._name = self._config['package']['name']
        self._version = self._config['package']['version']
        self._description = self._config['package']['description']
        self._authors = self._config['package']['authors']
        self._license = self._config['package'].get('license')
        self._homepage = self._config['package'].get('homepage')
        self._repository = self._config['package'].get('repository')
        self._keywords = self._config['package'].get('keywords', [])
        self._python_versions = self._config['package']['python']
        self._dependencies = self._get_dependencies(self._config.get('dependencies', {}))
        self._dev_dependencies = self._get_dependencies(self._config.get('dev-dependencies', {}))
        self._pip_dependencies = self._get_dependencies(self._config.get('dependencies', {}), 'pip')
        self._pip_dev_dependencies = self._get_dependencies(self._config.get('dev-dependencies', {}), 'pip')
        self._scripts = self._config.get('scripts', {})

        self._load_readme()

        self._include = self._config['package'].get('include', []) + list(self.INCLUDES)
        self._exclude = self._config['package'].get('exclude', []) + list(self.EXCLUDES)

        self._extensions = self._config.get('extensions', {})

    def _load_readme(self):
        readme = self._config['package']['readme']
        readme_path = os.path.join(self._dir, readme)

        if self.has_markdown_readme():
            if not pypandoc:
                warnings.warn(
                    'Markdown README files require the pandoc utility '
                    'and the pypandoc package.'
                )
            else:
                self._readme = pypandoc.convert_file(readme_path, 'rst')
        else:
            with open(readme_path) as f:
                self._readme = f.read()

    def has_markdown_readme(self):
        """
        Return whether the README is a markdown one.

        :rtype: boolean
        """
        readme = self._config['package']['readme']
        _, ext = os.path.splitext(readme)

        return ext == '.md'

    def is_lock(self):
        return False

    def build(self, **options):
        self.check()

        self._builder.build(self, **options)

    def check(self):
        """
        Checks if the poetry.toml file is valid.
        """
        package = self._config.get('package')
        if not package:
            raise MissingElement('package')

        self._check_package(package)

        dependencies = self._config.get('dependencies')
        if dependencies:
            self._check_dependencies(dependencies)

        dev_dependencies = self._config.get('dev-dependencies')
        if dev_dependencies:
            self._check_dependencies(dev_dependencies)

    def _check_package(self, package):
        if 'name' not in package:
            raise MissingElement('package.name')

        if 'version' not in package:
            raise MissingElement('package.version')

        authors = package.get('authors')
        if not authors:
            raise MissingElement('package.authors')

        if not isinstance(authors, list):
            raise InvalidElement('package.authors', 'it must be a list')

        license = package.get('license')
        if license:
            self._check_license(license)

        readme = package.get('readme')
        if not readme:
            raise MissingElement('package.readme')

        self._check_readme(readme)

    def _check_license(self, license):
        pass

    def _check_readme(self, readme):
        readme_path = os.path.join(self._dir, readme)

        if not os.path.exists(readme_path):
            raise InvalidElement('package.readme', 'invalid path provided')

        _, ext = os.path.splitext(readme)

        if ext not in ['.md', '.rst', '.txt']:
            raise InvalidElement('package.readme', 'extension [{}] is not supported'.format(ext))

    def _check_dependencies(self, dependencies):
        for name, constraint in dependencies.items():
            self._check_package_constraint(name, constraint)

    def _check_package_constraint(self, name, constraint):
        message = 'Invalid constraint [{}]'.format(constraint)

        if isinstance(constraint, dict):
            return self._check_vcs_constraint(name, constraint)
        else:
            try:
                return VersionParser().parse_constraints(constraint)
            except ValueError:
                pass

        raise InvalidElement('dependencies.{}'.format(name), message)

    def _check_vcs_constraint(self, name, constraint):
        if 'git' in constraint:
            self._check_git_constraint(name, constraint)

    def _check_git_constraint(self, name, constraint):
        if all(['branch' not in constraint, 'rev' not in constraint, 'tag' not in constraint]):
            raise InvalidElement(
                'dependencies.{}'.format(name),
                'Git constraint should have one of [branch, rev, tag]'
            )

    def _get_dependencies(self, dependencies, kind='default'):
        keys = sorted(list(dependencies.keys()))

        klass = Dependency
        if kind == 'pip':
            klass = PipDependency

        return [klass(k, dependencies[k]) for k in keys]



# -*- coding: utf-8 -*-

from .version_parser import VersionParser


class VersionSelector(object):

    def __init__(self, repository, parser=VersionParser()):
        self._repository = repository
        self._parser = parser

    def find_best_candidate(self, package_name, target_package_version=None,
                            preferred_stability='stable'):
        if target_package_version:
            constraint = self._parser.parse_constraints(target_package_version)
        else:
            constraint = None

        candidates = self._repository.find_packages(package_name, constraint)

        if not candidates:
            return False

        # Select highest version if we have many
        package = candidates[0]
        for candidate in candidates:
            # Select highest version of the two
            if package.version < candidate.version:
                package = candidate

        return package

    def find_recommended_require_version(self, package):
        version = package.version

        if not package.is_dev():
            return self._transform_version(version, package.pretty_version, package.stability)

    def _transform_version(self, version, pretty_version, stability):
        # attempt to transform 2.1.1 to 2.1
        # this allows you to upgrade through minor versions
        if str(version) != pretty_version:
            # Not semver
            return pretty_version

        # append stability flag if not stable
        if stability != 'stable':
            version += '@{}'.format(stability)

        return '^{}.{}'.format(version.major, version.minor)




# -*- coding: utf-8 -*-

from cleo import Application as BaseApplication

from ..version import VERSION
from .commands import (
    AboutCommand,
    CheckCommand,
    InitCommand,
    InstallCommand,
    LockCommand,
    PackageCommand,
    PublishCommand,
    RequireCommand,
    SearchCommand,
    UpdateCommand
)
from .commands.make import MakeSetupCommand


class Application(BaseApplication):
    """
    The console application that handles the commands.
    """

    def __init__(self):
        super(Application, self).__init__('Poet', VERSION)

    def get_default_commands(self):
        default_commands = super(Application, self).get_default_commands()

        return default_commands + [
            AboutCommand(),
            CheckCommand(),
            InitCommand(),
            InstallCommand(),
            LockCommand(),
            MakeSetupCommand(),
            PackageCommand(),
            PublishCommand(),
            RequireCommand(),
            SearchCommand(),
            UpdateCommand()
        ]



# -*- coding: utf-8 -*-

from ...version_selector import VersionSelector

from .command import Command


class RequireCommand(Command):
    """
    Add a dependency to the project.

    require
        {name* : Packages to add}
        {--dev : Packages should be added to dev-dependencies section}
        {--install : Install packages immediately}
    """

    def handle(self):
        packages = self.argument('name')
        is_dev = self.option('dev')
        install = self.option('install')
        if install:
            self.line('<warning>--install option is not supported yet.</>')

        requires = []

        for package in packages:
            constraint = None

            if ' ' in package:
                package, constraint = package.split(' ')

            matches = self._repository.search(package, 1)

            if not matches:
                self.line('<error>Unable to find package [{}]</>'.format(package))
                package = False
            else:
                exact_match = None
                choices = []

                for found_package in matches:
                    choices.append(found_package['name'])

                    if found_package['name'] == package:
                        exact_match = True
                        break

                if not exact_match:
                    self.line(
                        'Found <info>{}</info> packages matching <info>{}</info>'
                        .format(
                            len(matches),
                            package
                        )
                    )

                    package = self.choice(
                        '\nEnter package # to add, or the complete package name if it is not listed: ',
                        choices,
                        attempts=3
                    )

            # no constraint yet, determine the best version automatically
            if package is not False and ' ' not in package:
                self.line('')
                question = self.create_question(
                    'Enter the version constraint to require for <info>{}</> '
                    '(or leave blank to use the latest version): '
                    .format(package)
                )
                question.attempts = 3
                question.validator = lambda x: (x or '').strip() or False

                constraint = self.ask(question)

                if constraint is False:
                    constraint = self._find_best_version_for_package(package)

                    self.line(
                        'Using version <info>{}</info> for <info>{}</info>'
                        .format(constraint, package)
                    )

                package += ' {}'.format(constraint)

            if package is not False:
                requires.append(package)

        if requires:
            self.line('')
            section = '[dependencies]'
            if is_dev:
                section = '[dev-dependencies]'

            self.line('<comment>Add the following lines to the <info>{}</> section</>\n'.format(section))
            for require in requires:
                self.line('{} = "{}"'.format(*require.split(' ')))

    def _find_best_version_for_package(self, package):
        selector = VersionSelector(self._repository)
        package = selector.find_best_candidate(package)

        return selector.find_recommended_require_version(package)



# -*- coding: utf-8 -*-

from .about import AboutCommand
from .check import CheckCommand
from .init import InitCommand
from .install import InstallCommand
from .lock import LockCommand
from .package import PackageCommand
from .publish import PublishCommand
from .require import RequireCommand
from .search import SearchCommand
from .update import UpdateCommand



# -*- coding: utf-8 -*-

from ...installer import Installer

from .command import Command


class LockCommand(Command):
    """
    Lock the dependencies specified in <comment>poetry.toml</comment>.

    lock
        {--f|force : Force locking}
    """

    def handle(self):
        if self.has_lock() and not self.option('force'):
            return

        installer = Installer(self, self._repository)

        installer.lock(self.poet)



# -*- coding: utf-8 -*-

from pygments.lexer import RegexLexer
from pygments.token import (
    Text, Name, Number, String, Comment, Punctuation,
    Keyword, Operator
)


class TOMLLexer(RegexLexer):
    """
    Lexer for TOML, a simple language for config files
    """

    name = 'TOML'
    aliases = ['toml']
    filenames = ['*.toml']

    tokens = {
        'root': [

            # Basics, comments, strings
            (r'\s+', Text),
            (r'#.*?$', Comment.Single),
            (r'"(\\\\|\\"|[^"])*"', String),
            (r'(true|false)$', Keyword.Constant),
            ('[a-zA-Z_][a-zA-Z0-9_\-]*', Name),

            # Datetime
            (r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z', Number.Integer),

            # Numbers
            (r'(\d+\.\d*|\d*\.\d+)([eE][+-]?[0-9]+)?j?', Number.Float),
            (r'\d+[eE][+-]?[0-9]+j?', Number.Float),
            (r'\-?\d+', Number.Integer),

            # Punctuation
            (r'[]{}:(),;[]', Punctuation),
            (r'\.', Punctuation),

            # Operators
            (r'=', Operator)

        ]
    }



# -*- coding: utf-8 -*-

import os
import shutil

from .command import Command


class PackageCommand(Command):
    """
    Builds the package.

    package
        { --no-universal : Do not build a universal package. }
        { --no-wheels : Build only the source package. }
        { --c|clean : Make a clean package. }
    """

    def handle(self):
        poet = self.poet

        if self.option('clean'):
            egg_info = os.path.join(poet.base_dir, '{}.egg-info'.format(poet.name))

            if os.path.exists(egg_info):
                shutil.rmtree(egg_info)

        fmt = 'tar.gz'

        self.line('')
        self.line('<info>Building <comment>{}-{}</></>'.format(poet.name, poet.version))

        poet.build(universal=not self.option('no-universal'), no_wheels=self.option('no-wheels'))

        self.line('<info><comment>{}-{}</> built!</>'.format(poet.name, poet.version))
        self.line('')



# -*- coding: utf-8 -*-

from .command import Command


class CheckCommand(Command):
    """
    Check the <comment>poetry.toml</> file.

    check
    """

    def handle(self):
        self.line('')

        self.poet.check()

        self.info('The <comment>poetry.toml</> file is valid!')



# -*- coding: utf-8 -*-

import re

from semantic_version import Spec, Version


class VersionParser(object):

    _constraints = {}

    def parse_constraints(self, constraints):
        if not isinstance(constraints, list):
            constraints = constraints.replace(', ', ',')
        else:
            constraints = ','.join(constraints)

        if constraints not in self.__class__._constraints:
            specs = Spec(constraints)

            self.__class__._constraints[constraints] = specs

        return self.__class__._constraints[constraints]

    @classmethod
    def parse_stability(cls, version):
        if not isinstance(version, Version):
            version = Version.coerce(version)

        if version.prerelease:
            return 'dev'

        return 'stable'

    def parse_name_version_pairs(self, pairs):
        result = []

        for pair in pairs:
            pair = re.sub('^([^=: ]+)[=: ](.*)$', '\\1 \\2', pair.strip())

            if ' ' in pair:
                name, version = pair.split(' ', 2)

                result.append({
                    'name': name,
                    'version': version
                })
            else:
                result.append({
                    'name': pair,
                    'version': '*'
                })

        return result



# -*- coding: utf-8 -*-

import os

from twine.commands.upload import upload
from twine.commands.register import register
from requests.exceptions import HTTPError

from .command import Command


class PublishCommand(Command):
    """
    Publish the package to the remote repository.

    publish
        {--r|repository : The repository to register the package to}
    """

    def handle(self):
        # Checking if package exists
        dist = os.path.join(self.poet.base_dir, 'dist')
        package = os.path.join(dist, '{}-{}.tar.gz'.format(self.poet.name, self.poet.version))

        if not os.path.exists(package):
            self.info('No package found. Building it.')
            self.line('')
            self.call('package')
            self.line('')
            self.info('Package built.')
        else:
            self.info('Package found. Proceeding with publishing.')

        self.line('')
        self.line('Publishing <info>{}</> (<comment>{}</>)'.format(self.poet.name, self.poet.version))

        try:
            self._upload(dist, package)
        except HTTPError as e:
            if e.response.status_code != 403:
                raise

            # It may be the first time we publish the package
            # We'll try to register it and go from there
            try:
                self._register(package)
            except HTTPError:
                raise

            self._upload(dist, package)

    def _upload(self, dist, package):
        upload(
            [os.path.join(dist, '{}-{}*'.format(self.poet.name, self.poet.version))],
            self.option('repository') or 'pypi',
            False,
            None,
            None,
            None,
            None,
            None,
            '~/.pypirc',
            False,
            None,
            None,
            None
        )

    def _register(self, package):
        register(
            package,
            self.option('repository') or 'pypi',
            None, None, None,
            '~/.pypirc',
            None, None, None
        )



# -*- coding: utf-8 -*-

import os
import sys
import glob
import distutils

from cleo import Command as BaseCommand, InputOption

from ...repositories import PyPiRepository
from ...poet import Poet


class Command(BaseCommand):
    
    def __init__(self):
        super(Command, self).__init__()

        self._poet = None
        self._repository = PyPiRepository()
        self._virtual_env = None

    @property
    def poet_file(self):
        return os.path.join(os.getcwd(), 'poetry.toml')

    @property
    def poet(self):
        """
        Return the Poet instance.

        :rtype: poet.poet.Poet
        """
        if self._poet is None:
            self._poet = Poet(self.poet_file)

        return self._poet

    @property
    def lock_file(self):
        return self.poet.lock_file

    @property
    def virtual_env(self):
        return self._virtual_env

    def has_lock(self):
        return os.path.exists(self.lock_file)

    def configure(self):
        super(Command, self).configure()

        # Adding --i|index option
        self.add_option(
            'index', 'i',
            InputOption.VALUE_REQUIRED,
            'The index to use'
        )

    def execute(self, i, o):
        """
        Executes the command.
        """
        # Adding warning style
        self.set_style('warning', 'black', 'yellow')

        index = self.option('index')

        if index:
            self._repository = PyPiRepository(index)

        self.init_virtualenv()

        return super(Command, self).execute(i, o)

    def init_virtualenv(self):
        if 'VIRTUAL_ENV' not in os.environ:
            # Not in a virtualenv
            return

        # venv detection:
        # stdlib venv may symlink sys.executable, so we can't use realpath.
        # but others can symlink *to* the venv Python, so we can't just use sys.executable.
        # So we just check every item in the symlink tree (generally <= 3)
        p = os.path.normcase(sys.executable)
        paths = [p]
        while os.path.islink(p):
            p = os.path.normcase(os.path.join(os.path.dirname(p), os.readlink(p)))
            paths.append(p)

        p_venv = os.path.normcase(os.environ['VIRTUAL_ENV'])
        if any(p.startswith(p_venv) for p in paths):
            # Running properly in the virtualenv, don't need to do anything
            return

        if self.output.is_verbose():
            self.line('Using virtualenv <comment>{}</>'.format(os.environ['VIRTUAL_ENV']))

        if sys.platform == "win32":
            self._virtual_env = os.path.join(os.environ['VIRTUAL_ENV'], 'Lib', 'site-packages')
        else:
            lib = os.path.join(
                os.environ['VIRTUAL_ENV'], 'lib'
            )

            python = glob.glob(os.path.join(lib, 'python*'))[0].replace(lib + '/', '')

            self._virtual_env = os.path.join(
                lib,
                python,
                'site-packages'
            )

        import site
        sys.path.insert(0, self._virtual_env)

        site.addsitedir(self._virtual_env)

    def pip(self):
        if not self._virtual_env:
            return distutils.spawn.find_executable('pip')

        return os.path.realpath(
            os.path.join(self._virtual_env, '..', '..', '..', 'bin', 'pip')
        )



# -*- coding: utf-8 -*-

from ...installer import Installer

from .command import Command


class UpdateCommand(Command):
    """
    Update dependencies as according to the <comment>poetry.toml</> file.

    update
        { packages?* : The packages to update.}
    """

    def handle(self):
        if not self.has_lock():
            # We do not have a lock file
            # Assuming installation
            return self.call('install')

        installer = Installer(self, self._repository)

        installer.update(packages=self.argument('packages'))



# -*- coding: utf-8 -*-

from .package import PipDependency, Dependency
from .poet import Poet


class Lock(Poet):

    def is_lock(self):
        return True

    def load(self):
        root = self._config['root']
        self._name = root['name']
        self._version = root['version']

        packages = self._config['package']

        for package in packages:
            dep = Dependency(package['name'], package['version'])
            pip_dep = PipDependency(package['name'], package['version'], package.get('checksum'))

            if package['category'] == 'dev':
                self._dev_dependencies.append(dep)
                self._pip_dev_dependencies.append(pip_dep)
            else:
                self._dependencies.append(dep)
                self._pip_dependencies.append(pip_dep)




# -*- coding: utf-8 -*-

import os
import re

from setuptools import Extension
from setuptools.dist import Distribution
from pip.commands.wheel import WheelCommand
from semantic_version import Spec, Version

from .._compat import Path


SETUP_TEMPLATE = """# -*- coding: utf-8 -*-

from setuptools import setup

kwargs = dict(
    name={name},
    version={version},
    description={description},
    long_description={long_description},
    author={author},
    author_email={author_email},
    url={url},
    license={license},
    keywords={keywords},
    classifiers={classifiers},
    entry_points={entry_points},
    install_requires={install_requires},
    packages={packages},
    py_modules={py_modules},
    script_name='setup.py',
    include_package_data=True
)
"""

EXTENSIONS_TEMPLATE = """
from setuptools import Extension

kwargs['ext_modules'] = [
    {extensions}
]
"""

EXTENSION_TEMPLATE = """Extension(
        '{module}',
        {source}
    )"""


class Builder(object):
    """
    Tool to transform a poet file to a setup() instruction.
    """

    AUTHOR_REGEX = re.compile('(?u)^(?P<name>[- .,\w\d\'’"()]+) <(?P<email>.+?)>$')

    PYTHON_VERSIONS = {
        2: ['2.6', '2.7'],
        3: ['3.0', '3.1', '3.2', '3.3', '3.4', '3.5', '3.6']
    }

    def __init__(self):
        self._manifest = []

    def build(self, poet, **options):
        """
        Builds a poet.

        :param poet: The poet to build.
        :type poet: poet.poet.Poet
        """
        setup_kwargs = self._setup(poet, **options)

        setup = os.path.join(poet.base_dir, 'setup.py')
        manifest = os.path.join(poet.base_dir, 'MANIFEST.in')
        self._write_setup(setup_kwargs, setup)

        readme = None
        if poet.has_markdown_readme():
            readme = os.path.join(poet.base_dir, 'README.rst')
            if os.path.exists(readme):
                readme = None
            else:
                self._write_readme(readme, setup_kwargs['long_description'])

        self._manifest.append('include README.rst')

        self._write_manifest(manifest)

        try:
            dist = Distribution(setup_kwargs)
            dist.run_command('sdist')
        except Exception:
            raise
        finally:
            os.unlink(setup)
            os.unlink(manifest)

            if readme:
                os.unlink(readme)

        # Building wheel if necessary
        if not options.get('no_wheels'):
            command = WheelCommand()
            command_args = ['--no-index', '--no-deps', '--wheel-dir', 'dist', 'dist/{}'.format(poet.archive)]

            if options.get('universal', True):
                command_args.append('--build-option=--universal')

            command.main(command_args)

    def _setup(self, poet, **options):
        setup_kwargs = {
            'name': poet.name,
            'version': poet.version,
            'description': poet.description,
            'long_description': poet.readme,
            'include_package_data': True,
            'script_name': 'setup.py'
        }

        setup_kwargs.update(self._author(poet))

        setup_kwargs['url'] = self._url(poet)

        setup_kwargs['license'] = poet.license

        setup_kwargs['keywords'] = self._keywords(poet)

        setup_kwargs['classifiers'] = self._classifiers(poet)

        setup_kwargs['entry_points'] = self._entry_points(poet)

        setup_kwargs['install_requires'] = self._install_requires(poet)

        setup_kwargs.update(self._packages(poet))

        # Extensions
        setup_kwargs.update(self._ext_modules(poet))

        return setup_kwargs

    def _author(self, poet):
        m = self.AUTHOR_REGEX.match(poet.authors[0])

        return {
            'author': m.group('name'),
            'author_email': m.group('email')
        }

    def _url(self, poet):
        return poet.homepage or poet.repository

    def _keywords(self, poet):
        return ' '.join(poet.keywords or [])

    def _classifiers(self, poet):
        classifiers = []

        classifiers += self._classifiers_versions(poet)

        return classifiers

    def _classifiers_versions(self, poet):
        classifers = ['Programming Language :: Python']
        compatible_versions = {}

        for python in poet.python_versions:
            constraint = Spec(python)

            for major in [2, 3]:
                available_versions = self.PYTHON_VERSIONS[major]

                for version in available_versions:
                    if Version.coerce(version) in constraint:
                        if major not in compatible_versions:
                            compatible_versions[major] = []

                        compatible_versions[major].append(version)

        for major in sorted(list(compatible_versions.keys())):
            versions = compatible_versions[major]
            classifer_template = 'Programming Language :: Python :: {}'

            classifers.append(classifer_template.format(major))

            for version in versions:
                classifers.append(classifer_template.format(version))

        return classifers

    def _entry_points(self, poet):
        entry_points = {
            'console_scripts': []
        }

        for name, script in poet.scripts.items():
            entry_points['console_scripts'].append('{}={}'.format(name, script))

        return entry_points

    def _install_requires(self, poet):
        requires = []
        dependencies = poet.dependencies

        for dependency in dependencies:
            requires.append(dependency.normalized_name)

        return requires

    def _packages(self, poet):
        includes = poet.include
        packages = []
        modules = []
        crawled = []
        excluded = []

        for exclude in poet.exclude + poet.ignore:
            if not exclude:
                continue

            if exclude.startswith('/'):
                exclude = exclude[1:]

            for exc in Path().glob(exclude):
                if exc.suffix == '.py':
                    excluded.append('.'.join(exc.with_suffix('').parts))

        for include in includes:
            dirs = []
            others = []
            for element in Path().glob(include):
                if element.is_dir():
                    dirs.append(element)
                else:
                    others.append(element)

            m = re.match('^(.+)/\*\*/\*(\..+)?$', include)
            if m:
                # {dir}/**/* will not take the root directory
                # So we add it
                dirs.insert(0, Path(m.group(1)))

            for dir in dirs:
                package = '.'.join(dir.parts)

                if dir in crawled:
                    continue

                # We have a package
                if (dir / '__init__.py').exists():
                    children = [
                        c for c in dir.glob('*.py')
                    ]

                    filtered_children = [c for c in children if '.'.join(c.parts) not in excluded]
                    if children == filtered_children:
                        # If none of the children are excluded
                        # We have a full package
                        packages.append(package)
                    else:
                        modules += ['.'.join(c.parts) for c in filtered_children]

                    crawled += children

                crawled.append(dir)

            for element in others:
                if element in crawled or element.suffix == '.pyc':
                    continue

                if element.suffix == '.py' and element.name != '__init__.py':
                    modules.append('.'.join(element.with_suffix('').parts))
                elif element.suffix not in ['.py', '.pyc'] and '__pycache__' not in element.parts:
                    # Non Python file, add them to data
                    self._manifest.append('include {}\n'.format(element.as_posix()))
                elif element.name == '__init__.py':
                    dir = element.parent
                    children = [
                        c
                        for c in dir.glob('*.py')
                        if c.name != '__init__.py'
                    ]

                    if not children and dir not in crawled:
                        # We actually have a package
                        packages.append('.'.join(dir.parts))

                        crawled.append(dir)

                crawled.append(element)

        packages = [p for p in packages if p not in excluded]
        modules = [m for m in modules if m not in excluded]

        return {
            'packages': packages,
            'py_modules': modules
        }

    def _ext_modules(self, poet):
        extensions = []
        for module, source in poet.extensions.items():
            if not isinstance(source, list):
                source = [source]

            ext = Extension(module, source)
            extensions.append(ext)

        return {
            'ext_modules': extensions
        }

    def _write_setup(self, setup, dest):
        parameters = setup.copy()

        extensions = []
        if parameters['ext_modules']:
            for extension in parameters['ext_modules']:
                module = extension.name
                source = extension.sources

                extensions.append(
                    EXTENSION_TEMPLATE.format(
                        module=module,
                        source=repr(source)
                    )
                )

            del parameters['ext_modules']

        for key in parameters.keys():
            value = parameters[key]

            if value is not None:
                parameters[key] = repr(value)

        with open(dest, 'w') as f:
            f.write(SETUP_TEMPLATE.format(**parameters))

            if extensions:
                f.write(EXTENSIONS_TEMPLATE.format(extensions='\n    '.join(extensions)))

            f.write('\nsetup(**kwargs)')

    def _write_manifest(self, manifest):
        with open(manifest, 'w') as f:
            f.writelines(self._manifest)

    def _write_readme(self, readme, content):
        with open(readme, 'w') as f:
            f.write(content)




# -*- coding: utf-8 -*-

from pip.req import InstallRequirement
from .dependency import Dependency


class PipDependency(Dependency):

    def __init__(self, name, constraint, checksum=None):
        super(PipDependency, self).__init__(name, constraint)

        self._checksum = checksum

    @property
    def checksum(self):
        return self._checksum

    @property
    def normalized_name(self):
        normalized_name = self._name
        normalized_constraint = self.normalized_constraint

        if normalized_constraint:
            if self.is_vcs_dependency():
                normalized_name = normalized_constraint
            else:
                normalized_name += normalized_constraint

        return normalized_name

    def as_requirement(self):
        if self.is_vcs_dependency():
            return InstallRequirement.from_editable(self.normalized_name)

        return InstallRequirement.from_line(self.normalized_name)

    def _normalize_vcs_constraint(self, constraint):
        if 'git' in constraint:
            repo = constraint['git']
            if 'branch' in constraint:
                revision = constraint['branch']
            elif 'tag' in constraint:
                revision = constraint['tag']
            elif 'rev' in constraint:
                revision = constraint['rev']
            else:
                revision = 'master'

            if not repo.startswith('git+'):
                repo = 'git+' + repo

            return '{}@{}#egg={}'.format(repo, revision, self._name)

        raise ValueError('Unsupport VCS.')



# -*- coding: utf-8 -*-

import tempfile
import subprocess
import toml

import os
import shutil

from hashlib import sha256
from pip.download import unpack_url
from pip.index import Link
from piptools.resolver import Resolver
from piptools.repositories import PyPIRepository
from piptools.scripts.compile import get_pip_command
from piptools.cache import DependencyCache
from piptools.utils import is_pinned_requirement

from .locations import CACHE_DIR
from .package.pip_dependency import PipDependency


class Installer(object):

    def __init__(self, command, repository):
        self._command = command
        self._poet = command.poet
        self._repository = repository

    def install(self, dev=True):
        if not os.path.exists(self._poet.lock_file):
            self.lock(dev=dev)

            return self.install(dev=dev)

        self._command.line('')
        self._command.line('<info>Installing dependencies</>')
        self._command.line('')

        lock = self._poet.lock
        deps = lock.pip_dependencies

        if dev:
            deps += lock.pip_dev_dependencies

        for dep in deps:
            name = dep.name

            if dep.is_vcs_dependency():
                constraint = dep.pretty_constraint
            else:
                constraint = dep.normalized_constraint.replace('==', '')

            self._command.line('  - Installing <info>{}</> (<comment>{}</>)'.format(name, constraint))

            with open(os.devnull) as devnull:
                try:
                    subprocess.check_output([self._command.pip(), 'install', dep.normalized_name], stderr=devnull)
                except subprocess.CalledProcessError as e:
                    self._command.error('Error while installing [{}] ({})'.format(name, str(e)))
                    break

    def update(self, packages=None, dev=True):
        if self._poet.is_lock():
            raise Exception('Update is only available with a poetry.toml file.')

        self._command.line('')
        self._command.line('<info>Updating dependencies</>')
        self._command.line('')

        # Reading current lock
        lock = self._poet.lock
        current_deps = lock.pip_dependencies
        if dev:
            current_deps += lock.pip_dev_dependencies

        # Resolving new dependencies and locking them
        deps = self._poet.pip_dependencies
        if dev:
            deps += self._poet.pip_dev_dependencies

        if packages:
            packages = [p for p in self._resolve(deps) if p['name'] in packages]
        else:
            packages = self._resolve(deps)

        deps = [PipDependency(p['name'], p['version']) for p in packages]

        to_act_on = []
        for i, dep in enumerate(deps):
            action = None
            from_ = None
            found = False
            for j, current_dep in enumerate(current_deps):
                name = dep.name
                current_name = current_dep.name
                version = dep.normalized_constraint
                current_version = current_dep.normalized_constraint

                if name == current_name:
                    found = True

                    if version == current_version:
                        break

                    action = 'update'
                    from_ = current_dep
                    break

            if not found:
                action = 'install'

            if action:
                to_act_on.append((action, from_, dep))

        if not to_act_on:
            self._command.line('  - <info>Dependencies already up-to-date!</info>')

            return

        error = False
        for action, from_, dep in to_act_on:
            cmd = [self._command.pip(), 'install', dep.normalized_name]
            description = 'Installing'
            if action == 'update':
                description = 'Updating'
                cmd.append('-U')

            name = dep.name

            if dep.is_vcs_dependency():
                constraint = dep.pretty_constraint
            else:
                constraint = dep.normalized_constraint.replace('==', '')

            version = '<comment>{}</>'.format(constraint)

            if from_:
                if from_.is_vcs_dependency():
                    constraint = from_.pretty_constraint
                else:
                    constraint = from_.normalized_constraint.replace('==', '')

                version = '<comment>{}</> -> '.format(constraint) + version

            self._command.line('  - {} <info>{}</> ({})'.format(description, name, version))

            with open(os.devnull) as devnull:
                try:
                    subprocess.check_output(cmd, stderr=devnull)
                except subprocess.CalledProcessError as e:
                    error = True
                    self._command.error('Error while installing [{}] ({})'.format(name, str(e)))
                    break

        if not error:
            self._write_lock(packages)

    def lock(self, dev=True):
        if self._poet.is_lock():
            return

        self._command.line('')
        self._command.line('<info>Locking dependencies to <comment>poetry.lock</></>')
        self._command.line('')

        deps = self._poet.pip_dependencies

        if dev:
            deps += self._poet.pip_dev_dependencies

        packages = self._resolve(deps)

        self._write_lock(packages)

    def _resolve(self, deps):
        constraints = [dep.as_requirement() for dep in deps]

        command = get_pip_command()
        opts, _ = command.parse_args([])

        self._command.line('  - <info>Resolving dependencies</>')
        resolver = Resolver(
            constraints, PyPIRepository(opts, command._build_session(opts)),
            cache=DependencyCache(CACHE_DIR)
        )
        matches = resolver.resolve()
        pinned = [m for m in matches if not m.editable and is_pinned_requirement(m)]
        unpinned = [m for m in matches if m.editable or not is_pinned_requirement(m)]
        hashes = resolver.resolve_hashes(pinned)
        packages = []
        for m in matches:
            name = m.req.name
            version = str(m.req.specifier)
            if m in unpinned:
                url, specifier = m.link.url.split('@')
                rev, _ = specifier.split('#')

                version = self._get_vcs_version(url, rev)
                checksum = 'sha1:{}'.format(version['rev'])
            else:
                version = version.replace('==', '')
                checksum = list(hashes[m])

            package = {
                'name': name,
                'version': version,
                'checksum': checksum,
                'category': 'main'
            }

            packages.append(package)

        return sorted(packages, key=lambda p: p['name'].lower())

    def _get_vcs_version(self, url, rev):
        tmp_dir = tempfile.mkdtemp()
        current_dir = self._poet.base_dir

        try:
            unpack_url(Link(url), tmp_dir, download_dir=tmp_dir, only_download=True)

            os.chdir(tmp_dir)
            with open(os.devnull) as devnull:
                subprocess.check_output(['git', 'checkout', rev], stderr=devnull)

                revision = subprocess.check_output(['git', 'rev-parse', rev], stderr=devnull)

            revision = revision.decode().strip()
            version = {
                'git': url,
                'rev': revision
            }
        except Exception:
            raise
        finally:
            shutil.rmtree(tmp_dir)
            # Going back to current directory
            os.chdir(current_dir)

        return version

    def _write_lock(self, packages):
        self._command.line('  - <info>Writing dependencies</>')

        output = {
            'root': {
                'name': self._poet.name,
                'version': self._poet.version
            },
            'package': []
        }

        for package in packages:
            output['package'].append({
                'name': package['name'],
                'version': package['version'],
                'checksum': package.get('checksum'),
                'category': package['category']
            })

        with open(self._poet.lock_file, 'w') as f:
            f.write(toml.dumps(output))

    @classmethod
    def _hash(cls, filename):
        _hash = sha256()

        with open(filename, 'rb') as f:
            while True:
                chunk = f.read(8192)
                if not chunk:
                    break

                _hash.update(chunk)

        return _hash.hexdigest()



# -*- coding: utf-8 -*-


class PoetFileError(Exception):

    pass


class MissingElement(PoetFileError):

    def __init__(self, element):
        super(MissingElement, self).__init__(
            'The poetry.toml file is missing the [{}] element'.format(element)
        )


class InvalidElement(PoetFileError):

    def __init__(self, element, extra_info=''):
        msg = 'The element [{}] is invalid'.format(element)

        if extra_info:
            msg += ' ({})'.format(extra_info)

        super(InvalidElement, self).__init__(msg)



# -*- coding: utf-8 -*-

import os
import toml

from collections import OrderedDict
from pygments import highlight
from pygments.formatters.terminal import TerminalFormatter

from .command import Command
from ...version_parser import VersionParser
from ...version_selector import VersionSelector
from ...utils.lexers import TOMLLexer
from ...build import Builder


class InitCommand(Command):
    """
    Creates a basic <comment>poetry.toml</> file in current directory.

    init
        {--name= : Name of the package}
        {--description= : Description of the package}
        {--author= : Author name of the package}
        {--dependency=* : Package to require with a version constraint,
                          e.g. requests:^2.10.0 or requests==2.11.1}
        {--dev-dependency=* : Package to require for development with a version constraint,
                              e.g. requests:^2.10.0 or requests==2.11.1}
        {--l|license= : License of the package}
    """

    help = """
The <info>init</info> command creates a basic <comment>poetry.toml</> file
in the current directory.

<info>poet init</info>
"""
    def __init__(self):
        self._git_config = None

        super(InitCommand, self).__init__()

    def handle(self):
        formatter = self.get_helper('formatter')

        self.line([
            '',
            formatter.format_block('Welcome to the Poet config generator', 'bg=blue;fg=white', True),
            ''
        ])

        self.line([
            '',
            'This command will guide you through creating your <info>poetry.toml</> config.',
            ''
        ])

        poet_file = self.poet_file
        git_config = self.git_config()

        name = self.option('name')
        if not name:
            name = os.path.basename(os.path.dirname(poet_file))
            name = name.lower()

        question = self.create_question(
            'Package name [<comment>{}</comment>]: '
            .format(name),
            default=name
        )
        name = self.ask(question)

        version = '0.1.0'
        question = self.create_question(
            'Version [<comment>{}</comment>]: '.format(version),
            default=version
        )
        version = self.ask(question)

        description = self.option('description') or ''
        question = self.create_question(
            'Description [<comment>{}</comment>]: '
                .format(description),
            default=description
        )
        description = self.ask(question)

        author = self.option('author')

        if not author and git_config.get('user.name') and git_config.get('user.email'):
            author = '{} <{}>'.format(git_config['user.name'], git_config['user.email'])

        question = self.create_question(
            'Author [<comment>{}</comment>, n to skip]: '
                .format(author),
            default=author
        )
        question.validator = lambda v: self._validate_author(v, author)

        author = self.ask(question)

        license = self.option('license') or ''

        question = self.create_question(
            'License [<comment>{}</comment>]: '
                .format(license),
            default=license
        )

        license = self.ask(question)

        requirements = []

        question = '<question>Would you like to define your dependencies' \
                   ' (require) interactively [<comment>yes</comment>]?</>\n' \
                   '> '
        if self.confirm(question, True):
            requirements = self._format_requirements(
                self._determine_requirements(self.option('dependency'))
            )

        dev_requirements = []

        question = '<question>Would you like to define your dev dependencies' \
                   ' (require-dev) interactively [<comment>yes</comment>]?</>\n' \
                   '> '
        if self.confirm(question, True):
            dev_requirements = self._format_requirements(
                self._determine_requirements(self.option('dev-dependency'))
            )

        config = OrderedDict([
            ('package', OrderedDict([
                ('name', name)
            ]))
        ])

        output = toml.dumps(config)

        config = {'version': version}
        output += toml.dumps(config)

        config = {}
        if description:
            config['description'] = description
            output += toml.dumps(config)

        config = {}
        if author:
            config['authors'] = [author]
            output += toml.dumps(config)

        config = {}
        if license:
            config['license'] = license
            output += toml.dumps(config)

        output += '\n'

        if requirements:
            config = {}
            config['dependencies'] = requirements

            output += toml.dumps(config)
            output += '\n'

        if dev_requirements:
            config = {}
            config['dev-dependencies'] = dev_requirements

        output += toml.dumps(config) + '\n'

        if self.input.is_interactive():
            self.line('<info>Generated file</>')
            if self.output.is_decorated():
                self.line([
                    '',
                    highlight(
                        output,
                        TOMLLexer(),
                        TerminalFormatter()
                    ),
                    ''
                ])
            else:
                self.line(['', output, ''])

            if not self.confirm(
                'Do you confirm generation [<comment>yes</comment>]? ', True
            ):
                self.line('<error>Command aborted</error>')

                return 1

            with open(self.poet_file, 'w') as fd:
                fd.write(output)

    def _determine_requirements(self, requires):
        if requires:
            requires = self._normalize_requirements(requires)
            result = []

            for requirement in requires:
                if 'version' not in requirement:
                    # determine the best version automatically
                    version = self._find_best_version_for_package(requirement['name'])
                    requirement['version'] = version

                    self.line(
                        'Using version <info>{}</info> for <info{}</info>'
                        .format(requirement['version'], requirement['name'])
                    )

                result.append(requirement['name'] + ' ' + requirement['version'])

        version_parser = VersionParser()
        package = self.ask('Search for a package: ')
        while package is not None:
            matches = self._find_packages(package)

            if not matches:
                self.line('<error>Unable to find package</>')
                package = False
            else:
                exact_match = None
                choices = []

                for found_package in matches:
                    choices.append(found_package['name'])

                    if found_package['name'] == package:
                        exact_match = True
                        break

                if not exact_match:
                    self.line(
                        'Found <info>{}</info> packages matching <info>{}</info>'
                        .format(
                            len(matches),
                            package
                        )
                    )

                    package = self.choice(
                        '\nEnter package # to add, or the complete package name if it is not listed: ',
                        choices,
                        attempts=3
                    )

            # no constraint yet, determine the best version automatically
            if package is not False and ' ' not in package:
                question = self.create_question(
                    'Enter the version constraint to require '
                    '(or leave blank to use the latest version): '
                )
                question.attempts = 3
                question.validator = lambda x: (x or '').strip() or False

                constraint = self.ask(question)

                if constraint is False:
                    constraint = self._find_best_version_for_package(package)

                    self.line(
                        'Using version <info>{}</info> for <info>{}</info>'
                            .format(constraint, package)
                    )

                package += ' {}'.format(constraint)

            if package is not False:
                requires.append(package)

            package = self.ask('\nSearch for a package: ')

        return requires

    def _validate_author(self, author, default):
        author = author or default

        if author in ['n', 'no']:
            return

        m = Builder.AUTHOR_REGEX.match(author)
        if not m:
            raise ValueError(
                'Invalid author string. Must be in the format: '
                'John Smith <john@example.com>'
            )

        return author

    def _find_packages(self, package):
        return self._repository.search(package, 1)

    def _find_best_version_for_package(self, package):
        selector = VersionSelector(self._repository)
        package = selector.find_best_candidate(package)

        return selector.find_recommended_require_version(package)

    def _format_requirements(self, requirements):
        requires = OrderedDict()
        requirements = self._normalize_requirements(requirements)

        for requirement in requirements:
            requires[requirement['name']] = requirement['version']

        return requires

    def _normalize_requirements(self, requirements):
        parser = VersionParser()

        return parser.parse_name_version_pairs(requirements)

    def git_config(self):
        return self.poet.git_config



# -*- coding: utf-8 -*-

import os

from ....build import Builder

from ..command import Command


TEMPLATE="""from setuptools import setup

kwargs = {}

setup(**kwargs)
"""


class MakeSetupCommand(Command):
    """
    Renders a setup.py for testing purposes.

    make:setup
    """

    def handle(self):
        poet = self.poet
        builder = Builder()

        setup = builder._setup(poet)

        builder._write_setup(setup, os.path.join(os.path.dirname(self.poet_file), 'setup.py'))



# -*- coding: utf-8 -*-

import re

from ..version_parser import VersionParser, Version


class Package(object):

    def __init__(self, name, version):
        self._name = name
        self._version = Version.coerce(version) if not isinstance(version, Version) else version
        self._pretty_version = str(version)

        self._stability = VersionParser.parse_stability(version)
        self._dev = self._stability == 'dev'
        self._dependencies = []

    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version

    @property
    def pretty_version(self):
        return self._pretty_version

    @property
    def stability(self):
        return self._stability

    @property
    def dependencies(self):
        return self._dependencies

    @classmethod
    def check_name(cls, name):
        m = re.match('[a-z0-9\-]+/[a-z0-9\-]+', name)
        if m is None:
            return False

        return True

    def is_dev(self):
        return self._dev



# -*- coding: utf-8 -*-

from ...repositories import PyPiRepository
from .command import Command


class SearchCommand(Command):
    """
    Search for packages.

    search
        {tokens* : The tokens to search for.}
        {--N|only-name : Search only in name.}
    """

    help = """The search command searches for packages by its name
<info>poet search requests pendulum</info>
"""

    def handle(self):
        flags = PyPiRepository.SEARCH_FULLTEXT
        if self.option('only-name'):
            flags = PyPiRepository.SEARCH_NAME

        results = self._repository.search(self.argument('tokens'), flags)

        for result in results:
            self.line('')
            name = '<info>{}</>'.format(
                result['name']
            )

            if 'version' in result:
                name += ' (<comment>{}</>)'.format(result['version'])

            self.line(name)

            if 'description' in result:
                self.line(
                    ' {}'.format(result['description'])
                )



# -*- coding: utf-8 -*-

from ...installer import Installer

from .command import Command


class InstallCommand(Command):
    """
    Install and lock dependencies specified in the <comment>poetry.toml</comment> file.

    install
        {--no-dev : Do not install dev dependencies}
    """

    def handle(self):
        dev = not self.option('no-dev')

        installer = Installer(self, self._repository)

        installer.install(dev=dev)



# -*- coding: utf-8 -*-

from semantic_version import Spec, SpecItem, Version


class Dependency(object):

    def __init__(self, name, constraint):
        self._name = name
        self._constraint = constraint
        self._normalized_constraint = self._normalize(constraint)

    @property
    def name(self):
        return self._name

    @property
    def constraint(self):
        return self._constraint

    @property
    def pretty_constraint(self):
        constraint = self._constraint

        if self.is_vcs_dependency():
            if 'git' in self._constraint:
                vcs_kind = 'branch'
                if 'rev' in self._constraint:
                    vcs_kind = 'rev'
                    version = self._constraint['rev']
                elif 'tag' in constraint:
                    vcs_kind = 'tag'
                    version = self._constraint['tag']
                else:
                    version = self._constraint.get('branch', 'master')

                constraint = '{} {}'.format(vcs_kind, version)
            else:
                raise ValueError('Unsupport VCS.')

        return constraint

    @property
    def normalized_constraint(self):
        return self._normalized_constraint
    
    @property
    def normalized_name(self):
        normalized_name = self._name
        normalized_constraint = self.normalized_constraint

        if normalized_constraint:
            normalized_name += normalized_constraint

        return normalized_name

    def is_vcs_dependency(self):
        return isinstance(self._constraint, dict)

    def _normalize(self, constraint):
        if self.is_vcs_dependency():
            return self._normalize_vcs_constraint(constraint)

        constraint = Spec(constraint)
        normalized = []

        for spec in constraint.specs:
            major, minor, patch = spec.spec.major, spec.spec.minor, spec.spec.patch
            current = '{}.{}.{}'.format(major, minor or 0, patch or 0)
            current = Version(current)

            if spec.kind == SpecItem.KIND_CARET:
                if current.major != 0 or minor is None:
                    upper = current.next_major()
                elif current.minor != 0 or patch is None:
                    upper = current.next_minor()
                else:
                    upper = current.next_patch()

                normalized.append('>={},<{}'.format(current, upper))
            elif spec.kind == SpecItem.KIND_TILDE:
                if minor is None and patch is None:
                    upper = current.next_major()
                else:
                    upper = current.next_minor()

                upper = '{}.{}.{}'.format(*upper)

                normalized.append('>={},<{}'.format(current, upper))
            else:
                normalized.append(str(spec))

        return ','.join(normalized)

    def _normalize_vcs_constraint(self, constraint):
        # Neither setuptools nor distutils support VCS constraint
        # So by default we return nothing
        return

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.normalized_name)



# -*- coding: utf-8 -*-

import requests
from pip.models import PyPI

try:
    from xmlrpc.client import ServerProxy
except ImportError:
    from xmlrpclib import ServerProxy

from semantic_version import Version

from ..version_parser import VersionParser
from ..package import Package


class PyPiRepository(object):

    DEFAULT_URL = PyPI.pypi_url

    SEARCH_FULLTEXT = 0
    SEARCH_NAME = 1

    def __init__(self, url=DEFAULT_URL):
        self._url = url

    def find_packages(self, name, constraint=None):
        packages = []

        if constraint is not None:
            version_parser = VersionParser()
            constraint = version_parser.parse_constraints(constraint)

        with ServerProxy(self._url) as client:
            versions = client.package_releases(name, True)

        if constraint:
            versions = constraint.select([Version.coerce(v) for v in versions])

        for version in versions:
            try:
                packages.append(Package(name, version))
            except ValueError:
                continue

        return packages

    def search(self, query, mode=0):
        results = []

        search = {
            'name': query
        }

        if mode == self.SEARCH_FULLTEXT:
            search['summary'] = query

        client = ServerProxy(self._url)
        hits = client.search(search, 'or')

        for hit in hits:
            results.append({
                'name': hit['name'],
                'description': hit['summary'],
                'version': hit['version']
            })

        return results

    def package_name(self, name):
        url = 'https://pypi.python.org/pypi/{}/json'.format(name)

        response = requests.get(url)

        if response.status_code == 404:
            raise Exception('Package [{}] not found'.format(name))

        response.raise_for_status()

        info = response.json()['info']

        return info['name']



# -*- coding: utf-8 -*-

from cleo import Command


class AboutCommand(Command):
    """
    Short information about Poet.

    about
    """

    def handle(self):
        self.line("""<info>Poet - Package Management for Python</info>

<comment>Poet is a dependency manager tracking local dependencies of your projects and libraries.
See <fg=blue>https://github.com/sdispater/poet</> for more information.</comment>
""")
